import sys

from django.conf import settings as django_settings
from django.core.exceptions import ImproperlyConfigured


class Settings(object):
    """
    Reads and validates django_qc settings.
    """

    def __init__(self) -> None:
        self.RAISE_EXC = True
        self.LOG_EXC = True
        self.DEBUG = True

        if hasattr(django_settings, 'DB_HELPER'):
            db_helper_settings = django_settings.DB_HELPER
            for setting, value in db_helper_settings.items():
                if hasattr(self, setting):
                    setattr(self, setting, value)
                else:
                    raise ImproperlyConfigured(f'`{setting}` is not a valid setting for DB_HELPER')

        if self.DEBUG and not django_settings.DEBUG and 'test' not in sys.argv:
            raise ImproperlyConfigured('DB-helper should only run when DEBUG is True')


settings = Settings()



import sys
from unittest.mock import patch

from django.core.exceptions import ImproperlyConfigured
from django.test import SimpleTestCase

from django_qc.settings import Settings


class TestSettings(SimpleTestCase):
    """
    Tests package settings.
    """

    @patch('django_qc.settings.django_settings')
    def test_excess_input(self, setting):
        """
        Make sure we alert the user about excess settings, as this is most likely a mistake on their part.
        """
        setting.DB_HELPER = {'TEST': 'test'}
        with self.assertRaisesRegexp(ImproperlyConfigured, '`TEST` is not a valid setting for DB_HELPER'):
            Settings()

    @patch('django_qc.settings.django_settings')
    def test_debug_false(self, setting):
        """
        Make sure we alert the user when trying to run the program if DEBUG is false.
        """
        setting.DEBUG = False
        if 'test' in sys.argv:
            sys.argv.pop(sys.argv.index('test'))
        with self.assertRaisesRegexp(ImproperlyConfigured, 'DB-helper should only run when DEBUG is True'):
            Settings()



# Generated by Django 3.0.4 on 2020-03-15 14:25

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='CD',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50)),
            ],
        ),
        migrations.CreateModel(
            name='Song',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50)),
                ('cd', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tests.CD')),
            ],
        ),
    ]



from django.conf import settings
from django.test import TestCase

from django_qc.core import db_helper
from tests.models import CD, Song


class TestingBaseClass(TestCase):

    @classmethod
    def setUpTestData(cls):
        cd = CD.objects.create(name='myAlbum')
        for i in range(12):
            Song.objects.create(name=f'song-{i}', cd=cd)

    def setUp(self) -> None:
        settings.DEBUG = True

    def call(self, count):
        @db_helper(count=count)
        def call_db():
            songs = Song.objects.all()
            for song in songs:
                _ = (song.name, song.cd)

        call_db()



from unittest.mock import patch

from django.core.exceptions import ImproperlyConfigured

from django_qc.core import db_helper
from tests.base import TestingBaseClass
from tests.models import Song


class TestWrapper(TestingBaseClass):

    def test_incorrect_count(self):
        """
        Test that an n+1 error is caught and error is raised.
        """
        with self.assertRaisesRegexp(ImproperlyConfigured, '`call_db` performed 13 queries, where we expected 1'):
            self.call(1)

    @patch('django_qc.core.settings.RAISE_EXC', False)
    def test_correct_error_logging(self):
        """
        An exception logger should be logged if RAISE_EXC is False.
        """
        with self.assertLogs(level='ERROR') as log:
            self.call(1)
            self.assertIn('Function `call_db` performed 13 queries, where we expected 1', log.output[0])

    @patch('django_qc.core.settings.RAISE_EXC', False)
    @patch('django_qc.core.settings.LOG_EXC', False)
    def test_warning_logging(self):
        """
        One warning logger should be raised if both settings are False.
        """
        with self.assertLogs(level='WARNING') as log:
            self.call(1)
            self.assertIn('Function `call_db` performed 13 queries, where we expected 1', log.output[0])
            self.assertEqual(1, len(log.output))  # If more than 1, logging is incorrectly set up

    def test_exceeded_maximum_count(self):
        """
        Verify that an error is raised if count exceeds the max.
        """
        with self.assertRaisesRegexp(ImproperlyConfigured, 'Count cannot be higher than the max query limit of 9000'):
            @db_helper(count=9001)
            def _():
                pass

    def test_correct_count(self):
        """
        Good DB call.
        """
        @db_helper(count=1)
        def call_db_well():
            songs = Song.objects.all().prefetch_related('cd')
            for song in songs:
                _ = (song.name, song.cd)



__version__ = '0.2.1'
__author__ = 'Sondre Lillebø Gundersen'

from setuptools import setup, find_packages


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('CHANGELOG.rst') as changelog_file:
    changelog = changelog_file.read()

setup(
    name='django-qc',
    version=__version__,
    description='DB utility to help you catch query inefficiencies in Django.',
    py_modules=['django_qc'],
    include_package_data=True,
    long_description=readme + '\n\n' + changelog,
    license='BSD',
    author=__author__,
    author_email='sondrelg@live.no',
    url='https://github.com/sondrelg/django-query-counter',
    download_url='https://pypi.org/project/django-qc',
    packages=find_packages(exclude=['']),
    install_requires=['django'],
    keywords=['orm', 'database', 'queries', 'testing', 'performance'],
    platforms='OS Independent',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Documentation',
    ],
)



import logging
from functools import wraps
from typing import Callable

from django.core.exceptions import ImproperlyConfigured
from django.db import connections
from django.db.backends.base.base import BaseDatabaseWrapper

from django_qc.settings import settings

logger = logging.getLogger('django_qc')


def db_helper(count: int):
    """
    Decorator function to count DB queries.
    """
    # If the intended count of queries is higher than the max, connect.queries will reset.
    if count > BaseDatabaseWrapper.queries_limit:
        raise ImproperlyConfigured(f'Count cannot be higher than the max query limit of {BaseDatabaseWrapper.queries_limit}')

    def outer(fn: Callable):
        """
        Wrapper function - outer layer.
        """

        @wraps(fn)
        def inner(*args, **kwargs):
            query_count = sum(len(connections[db_name].queries) for db_name in connections)
            fn(*args, **kwargs)
            actual_count = sum(len(connections[db_name].queries) for db_name in connections) - query_count

            # If the amount of queries exceeds expectations
            if settings.DEBUG and actual_count != count:
                error_msg = f'Function `{fn.__name__}` performed {actual_count} queries, where we expected {count}'
                if settings.RAISE_EXC:
                    raise ImproperlyConfigured(error_msg)
                elif settings.LOG_EXC:
                    logger.exception(error_msg)
                else:
                    logger.warning(error_msg)

        return inner

    return outer



from django.db import models


class CD(models.Model):
    name = models.CharField(max_length=50)


class Song(models.Model):
    name = models.CharField(max_length=50)
    cd = models.ForeignKey(CD, on_delete=models.CASCADE, blank=False, null=False)
